#!/usr/bin/env python
# coding: utf-8
import numpy as np
import pandas as pd
import rqdatac as rq
import datetime
import traceback
import json
from ...wrapper.mysql import RawDatabaseConnector
from ...view.raw_models import *
from .constant import STOCK_VALUATION_FACTORS, FUND_INDICATORS
from ...api.basic import BasicDataApi
from .raw_data_helper import RawDataHelper

class RqRawDataDownloader(object):
    def __init__(self, rq_license, data_helper):
        self._data_helper = data_helper
        # Initialize RiceQuant API
        rq.init('license', rq_license, ('rqdatad-pro.ricequant.com', 16011))

        # Get all fund order_book_ids
        # self._fund_order_book_id_list = list(set(rq.fund.all_instruments()['order_book_id'].tolist()))

        # Get stock order_book_ids
        # self._stock_order_book_id_list = rq.all_instruments(type='CS', market='cn')['order_book_id'].tolist()

        # Get index list
        # index_df = BasicDataApi().get_index_info()
        # self._index_list = [order_book_id for order_book_id in index_df['order_book_id'].tolist()
        #     if order_book_id and order_book_id != 'not_available' and '.' in order_book_id]
        # self._future_list = [order_book_id for order_book_id in index_df['order_book_id'].tolist()
        #     if order_book_id and order_book_id != 'not_available' and '.' not in order_book_id]
        # self._order_book_id_2_index_id = {order_book_id : index_id for order_book_id, index_id in
        #     zip(index_df['order_book_id'], index_df['index_id']) if order_book_id != 'not_available'}

    def _get_date_list(self, start_date, end_date):
        start_datetime = datetime.datetime.strptime(start_date, '%Y%m%d')
        end_datetime = datetime.datetime.strptime(end_date, '%Y%m%d')
        date_list = []
        while start_datetime <= end_datetime:
            date_list.append(start_datetime.strftime('%Y%m%d'))
            start_datetime += datetime.timedelta(days=1)
        return date_list

    def rq_fund_indicator(self, start_date, end_date):
        try:
            df = rq.fund.get_indicators(self._fund_order_book_id_list, start_date, end_date,
                fields=FUND_INDICATORS)

            if df is not None:
                df.reset_index(inplace=True)
                self._data_helper._upload_raw(df, RqFundIndicator.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_fund_nav(self, start_date, end_date):
        try:
            df_list = []

            count = 0
            for order_book_id in self._fund_order_book_id_list:
                count += 1
                if count % 100 == 0:
                    print(f'[{RqFundNav.__table__.name}] downloaded {count} out of '
                        f'{len(self._fund_order_book_id_list)} funds')
                df = rq.fund.get_nav(order_book_id, start_date, end_date, expect_df=True)
                if df is None:
                    continue
                df.reset_index(inplace=True)
                df_list.append(df)

            if len(df_list) > 0:
                df_all = pd.concat(df_list)
                self._data_helper._upload_raw(df_all, RqFundNav.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_fund_size(self):
        try:
            instruments = []
            for instrument in rq.fund.instruments(self._fund_order_book_id_list):
                instruments.append({
                    'order_book_id' : instrument.order_book_id,
                    'latest_size' : instrument.latest_size
                })
            df = pd.DataFrame(instruments)
            if df is not None:
                self._data_helper._upload_raw(df, RqFundSize.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_index_price(self, start_date, end_date):
        try:
            df = rq.get_price(self._index_list, start_date, end_date, expect_df=True)
            if df is not None:
                df = df.reset_index().drop(['num_trades'], axis = 1)
                df['datetime'] = df['date']
                df = df.drop(['date'], axis = 1)
                self._data_helper._upload_raw(df, RqIndexPrice.__table__.name)

            df = rq.get_price(self._future_list, start_date, end_date, expect_df=True)
            if df is not None:
                df = df.reset_index().drop(['limit_down', 'prev_settlement', 'limit_up', 'open_interest', 'settlement', 
                    'dominant_id'], axis = 1, errors='ignore')
                df['datetime'] = df['date']
                df = df.drop(['date'], axis = 1)
                self._data_helper._upload_raw(df, RqIndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_index_weight(self, start_date, end_date):
        try:
            date_list = self._get_date_list(start_date, end_date)
            for i in self._index_list:
                res = []
                for d in date_list:
                    try:
                        df = rq.index_weights(i, date=d)
                        if df is not None:
                            dic = df.to_dict()
                            stock_list = [k for k, v in dic.items()]
                            weight_list = [round(v, 6) for k, v in dic.items()]
                            res.append({
                                'index_id': self._order_book_id_2_index_id[i],
                                'datetime': d,
                                'weight_list': json.dumps(weight_list),
                                'stock_list': json.dumps(stock_list),
                            })
                    except:
                        pass
                if len(res) > 0:
                    df = pd.DataFrame(res)
                    self._data_helper._upload_raw(df, RqIndexWeight.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_stock_fin_fac(self, start_date, end_date):
        try:
            factor = 'return_on_equity_ttm'
            df = rq.get_factor(self._stock_order_book_id_list, factor=factor, start_date=start_date, end_date=end_date)
            if df is not None:
                to_convert_series = type(df) is pd.core.series.Series
                if to_convert_series:
                    df = df.to_frame()
                df = df.stack().reset_index()
                df.columns = ['stock_id', 'datetime', factor] if to_convert_series else ['datetime', 'stock_id', factor]
                self._data_helper._upload_raw(df, RqStockFinFac.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_stock_post_price(self, start_date, end_date):
        try:
            df = rq.get_price(self._stock_order_book_id_list, start_date=start_date, end_date=end_date,
                adjust_type='post', expect_df=True)
            if df is not None:
                df = df.reset_index().rename(columns={'date': 'datetime'})
                self._data_helper._upload_raw(df, RqStockPostPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_stock_price(self, start_date, end_date):
        try:
            df = rq.get_price(self._stock_order_book_id_list, start_date=start_date, end_date=end_date,
                adjust_type='none', expect_df=True)
            if df is not None:
                df = df.reset_index().rename(columns={'date': 'datetime'})
                self._data_helper._upload_raw(df, RqStockPrice.__table__.name)
            return True
        except:
            traceback.print_exc()
            return False

    def rq_stock_valuation(self, start_date, end_date):
        try:
            df = rq.get_factor(self._stock_order_book_id_list, STOCK_VALUATION_FACTORS, start_date=start_date,
                end_date=end_date, expect_df=True)
            if df is not None:
                df = df.reset_index().rename(columns={'order_book_id': 'stock_id', 'date':'datetime'}).replace(
                    [np.inf, -np.inf], np.nan)
                self._data_helper._upload_raw(df, RqStockValuation.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def trading_day_list(self, start_date, end_date):
        try:
            df = pd.DataFrame(rq.get_trading_dates(start_date=start_date, end_date=end_date, market='cn'),
                columns=['datetime'])
            if df is not None:
                self._data_helper._upload_raw(df, TradingDayList.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def stock_info(self):
        try:
            order_book_ids = []
            for stock_order_book_id in self._stock_order_book_id_list:
                order_book_ids.append({
                    'rq_id' : stock_order_book_id,
                    'stock_id' : stock_order_book_id
                })
            df = pd.DataFrame(order_book_ids)
            if df is not None:
                self._data_helper._upload_raw(df, StockInfo.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_rating(self, start_date, end_date):
        try:
            fund_order_book_id_list = list(set(rq.fund.all_instruments()['order_book_id'].tolist()))
            df = rq.fund.get_ratings(fund_order_book_id_list)
            # df = rq.fund.get_ratings(self._fund_order_book_id_list)
            if df is not None:
                df.reset_index(inplace=True)
                df = df[(df['datetime'] >= start_date) & (df['datetime'] <= end_date)]
                self._data_helper._upload_raw(df, FundRating.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def stock_minute(self):
        try:
            stock_order_book_id_list = rq.all_instruments(type='CS', market='cn')['order_book_id'].tolist()
            df = rq.current_minute(stock_order_book_id_list, skip_suspended=True)
            if df is not None:
                today_date = datetime.date.today()
                today = datetime.datetime(today_date.year, today_date.month, today_date.day)
                df.reset_index(inplace=True)
                df = df[df['datetime'] >= today]
                self._data_helper._upload_raw(df, RqStockMinute.__table__.name)
            return df
        except Exception as e:
            print(e)
            traceback.print_exc()
            return None

    def index_minute(self):
        # Get index list
        try:
            index_df = BasicDataApi().get_index_info()
            index_list = [order_book_id for order_book_id in index_df['order_book_id'].tolist()
                if order_book_id and order_book_id != 'not_available' and '.' in order_book_id]
            df = rq.current_minute(index_list, skip_suspended=True)
            if df is not None:
                today = datetime.date.today()
                df.reset_index(inplace=True)
                df = df[df['datetime'] >= today]
                # self._data_helper._upload_raw(df, RqIndexMinute.__table__.name)
            print(df)
            return df
        except Exception as e:
            print(e)
            traceback.print_exc()
            return None

    def download_all(self, start_date, end_date):
        failed_tasks = []
        # # Safe to disable
        # if not self.rq_fund_indicator(start_date, end_date):
        #     failed_tasks.append('rq_fund_indicator')

        # # Safe to disable
        # if not self.rq_fund_nav(start_date, end_date):
        #     failed_tasks.append('rq_fund_nav')

        # # Safe to disable
        # if not self.rq_index_weight(start_date, end_date):
        #     failed_tasks.append('rq_index_weight')

        # # Safe to disable
        # if not self.rq_stock_fin_fac(start_date, end_date):
        #     failed_tasks.append('rq_stock_fin_fac')

        # # Safe to disable
        # if not self.rq_stock_post_price(start_date, end_date):
        #     failed_tasks.append('rq_stock_post_price')

        # # Safe to disable
        # if not self.rq_stock_price(start_date, end_date):
        #     failed_tasks.append('rq_stock_price')

        # # Safe to disable
        # if not self.rq_stock_valuation(start_date, end_date):
        #     failed_tasks.append('rq_stock_valuation')

        # # Safe to disable
        # if not self.stock_info():
        #     failed_tasks.append('stock_info')

        # # Safe to disable
        # if not self.trading_day_list(start_date, end_date):
        #     failed_tasks.append('trading_day_list')

        # # Safe to disable
        # if not self.rq_fund_size():
        #     failed_tasks.append('rq_fund_size')

        # # Safe to disable
        # if not self.rq_index_price(start_date, end_date):
        #     failed_tasks.append('rq_index_price')

        # TODO: used in basic db, replace to EM
        if not self.fund_rating(start_date, end_date):
            failed_tasks.append('fund_rating')

        return failed_tasks


if __name__ == "__main__":
    from ....util.config import SurfingConfigurator

    rq_license = SurfingConfigurator().get_license_settings('rq')
    raw_data_downloader = RqRawDataDownloader(rq_license, RawDataHelper())
    raw_data_downloader.stock_minute()
    # raw_data_downloader.rq_index_price('20200421', '20200422')
    # raw_data_downloader.rq_index_weight('20200428', '20200428')
    # raw_data_downloader.rq_stock_price('20050101', '20091231')




from typing import Tuple, Set

from collections import defaultdict
from ...wrapper.mysql import RawDatabaseConnector


class RawDataHelper(object):
    def __init__(self):
        self._updated_count = defaultdict(int)

    def _upload_raw(self, df, table_name, to_truncate=False):
        print(table_name)
        # print(df)
        if to_truncate:
            with RawDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
        df.to_sql(table_name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')

        self._updated_count[table_name] += df.shape[0]

    @staticmethod
    def get_new_and_delisted_fund_list(date: str) -> Tuple[Set[str], Set[str]]:
        import pandas as pd
        from ...api.raw import RawDataApi

        fund_list = RawDataApi().get_em_fund_list(date, limit=2)
        if fund_list.shape[0] < 2:
            print(f"Failed to get two days' fund list, (date){date} (len){fund_list.shape[0]}")
            return (None, None)

        lastest_list: pd.Series = fund_list.iloc[0, :]
        last_list: pd.Series = fund_list.iloc[1, :]
        print(f"got two days' fund list, (latest){lastest_list.datetime} (prev){last_list.datetime}")
        return (set(lastest_list.all_live_fund_list.split(',')).difference(set(last_list.all_live_fund_list.split(','))),
                set(lastest_list.delisted_fund_list.split(',')).difference(set(last_list.delisted_fund_list.split(','))))



from typing import List, Dict, Optional
import pandas as pd
import numpy as np
import datetime
import traceback
import json
from functools import partial
import math
from sklearn.metrics import r2_score
import statsmodels.api as sm
import concurrent.futures
from sklearn.linear_model import LinearRegression
from scipy.optimize import Bounds, minimize
from statsmodels.tsa.ar_model import AutoReg, ar_select_order
from ...manager.manager_fund import FundDataManager
from ...manager.score import FundScoreManager
from ...wrapper.mysql import DerivedDatabaseConnector
from ...api.basic import BasicDataApi

class FundIndicatorProcessorGroup(object):

    _TRADING_DAYS_PER_YEAR = 242
    _NATURAL_DAYS_PER_YEAR = 365
    _RISK_FEE_RATE = 0.025
    _RISK_FEE_RATE_PER_DAY = _RISK_FEE_RATE / _TRADING_DAYS_PER_YEAR

    def __init__(self, data_helper):
        self._data_helper = data_helper
        self._basic_api = BasicDataApi()

    def init(self, start_date='20200809', end_date='20200809'):    
        self.log = []
        start_date: str = (pd.to_datetime(start_date) - datetime.timedelta(days=self._NATURAL_DAYS_PER_YEAR*2+1)).strftime('%Y%m%d')
        # 获取区间内交易日列表
        all_trading_days: pd.Series = self._basic_api.get_trading_day_list().drop(columns='_update_time').set_index('datetime')
        self._trading_days: pd.Series = all_trading_days.loc[pd.to_datetime(start_date).date():pd.to_datetime(end_date).date()]
        self.end_date = self._trading_days.index.values[-1]

        # 基金净值数据、指数价格数据的index对齐到交易日列表
        fund_nav: pd.DataFrame = self._basic_api.get_fund_nav_with_date(start_date, end_date)#, fund_list = f_l)
        self._fund_nav: pd.DataFrame = fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(self._trading_days.index).fillna(method='ffill')
        # 选取成立满半年的
        res = []
        for fund_id in self._fund_nav:
            if self._fund_nav[fund_id].dropna().shape[0] > 120:
                res.append(fund_id)
        self._fund_nav = self._fund_nav[res].copy()

        index_price: pd.DataFrame = self._basic_api.get_index_price_dt(start_date, end_date).drop(columns='_update_time')
        # 有的index最近没有数据，如活期存款利率/各种定期存款利率等，需要先reindex到全部的交易日列表上ffill
        self._index_price = index_price.pivot_table(index='datetime', columns='index_id', values='close').reindex(self._trading_days.index)#.fillna(method='ffill')
        _bank_rate_df = self._basic_api.get_index_price(index_list =['ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d']).drop(columns='_update_time')
        _bank_rate_df = _bank_rate_df.pivot_table(index='datetime', columns='index_id', values='close').reindex(all_trading_days.index).fillna(method='ffill').reindex(self._trading_days.index)
        self._index_price = pd.concat([self._index_price, _bank_rate_df], axis=1)
        # 再reindex到参与计算的交易日列表上
        #self._index_price: pd.DataFrame = index_price.reindex(self._trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._index_price.index)
        try:
            # 这个指数有一天的点数是0，特殊处理一下
            self._index_price['spi_spa'] = self._index_price['spi_spa'].where(self._index_price.spi_spa != 0).fillna(method='ffill')
        except KeyError:
            pass

        # 对数净值取差分并且去掉第一个空值，得到基金对数收益率数列
        self._fund_ret: pd.DataFrame = np.log(self._fund_nav).diff().iloc[1:, :]

        self._fund_size = self._basic_api.get_fund_size_range(start_date, end_date)
        self._fund_size = self._fund_size.pivot_table(index='datetime', columns='fund_id', values='size').reindex(self._fund_ret.index).fillna(method='ffill')
        
        # 计算benchmark return，且index对齐到fund_ret
        benchmark_ret: pd.DataFrame = self._get_benchmark_return()
        self._benchmark_ret: pd.DataFrame = benchmark_ret.reindex(self._fund_ret.index)
        pd.testing.assert_index_equal(self._fund_ret.index, self._benchmark_ret.index)

        # 获取待计算的基金列表
        fund_list: pd.DataFrame = self._basic_api.get_fund_info().drop(columns='_update_time')
        self._fund_list: pd.DataFrame = fund_list[fund_list.structure_type <= 1]
        # 获取wind一级分类
        self._wind_class_1: np.ndarray = fund_list.wind_class_1.unique()
        self.get_continue_stats_tool()

    def df_resample(self, df):
        df.index = pd.to_datetime(df.index)
        df = df.resample('1W').sum()
        df.index = pd.Series(df.index).dt.date
        return df

    def stutzer_index(self, ex_return):
        ex_return = np.array(ex_return)
        information_statistic = lambda theta: np.log(np.mean(np.exp(theta[0] * ex_return)))
        theta0 = [-1.]
        bounds = Bounds((-np.inf), (100))
        result = minimize(information_statistic, theta0, method='SLSQP', bounds=bounds, tol=1e-16)
        if result.success:
            information_statistic = -result.fun
            stutzer_index = np.abs(np.mean(ex_return)) / np.mean(ex_return) * np.sqrt(2 * information_statistic)
            result = {'information_statistic': information_statistic,
                    'stutzer_index': stutzer_index,
                    'theta': result.x[0]}
        else:
            result = {'information_statistic': information_statistic,
                    'stutzer_index': np.nan,
                    'theta': result.x[0]}
            #print('未找到Information Statistic最大值')
            return result
        return result

    def _lambda_cl(self, x: pd.Series, fund_ret: pd.DataFrame):
        fund_id = x.name
        total = pd.concat({'Y': fund_ret[fund_id], 'x': x}, axis=1)
        total = total[total.notna().all(axis=1)]
        if total.empty:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        Y: pd.Series = total['Y']
        x = total['x']
        if x.count() != Y.count():
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X: pd.DataFrame = pd.concat([x, x], axis=1)
        X.columns = [0, 1]
        X[0][X[0] < 0] = 0
        X[1][X[1] > 0] = 0
        X['const'] = 1
        X2 = X.copy()
        est = sm.OLS(Y, X2)
        est2 = est.fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params.const,
                'alpha_t':est2.tvalues.const,
                'alpha_p':est2.pvalues.const,
                # TODO beta 2 - beta 1 , p values and t values calculation is not correct
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                } 

    def _lambda_tm(self, x: pd.Series, fund_ret: pd.DataFrame):
        fund_id = x.name
        total = pd.concat({'Y': fund_ret[fund_id], 'x': x}, axis=1)
        total = total[total.notna().all(axis=1)]
        if total.empty:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        Y: pd.Series = total['Y']
        x = total['x']
        if x.count() != Y.count():
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X: pd.DataFrame = pd.concat([x, x*x], axis=1)
        X.columns = [0, 1]
        X['const'] = 1
        X2 = X.copy()
        est = sm.OLS(Y, X2)
        est2 = est.fit()
        return {'beta':est2.params[1],
                'alpha':est2.params.const,
                'alpha_t':est2.tvalues.const,
                'alpha_p':est2.pvalues.const,
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }  

    def _lambda_hm(self, x: pd.Series, fund_ret: pd.DataFrame):
        fund_id = x.name
        total = pd.concat({'Y': fund_ret[fund_id], 'x': x}, axis=1)
        total = total[total.notna().all(axis=1)]
        if total.empty:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        Y: pd.Series = total['Y']
        x = total['x']
        if x.count() != Y.count():
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X: pd.DataFrame = pd.concat([x, x], axis=1)
        X.columns = [0, 1]
        X[1][X[1] < 0] = 0
        X['const'] = 1
        X2 = X.copy()
        est = sm.OLS(Y, X2)
        est2 = est.fit()
        return {'beta':est2.params[1],
                'alpha':est2.params.const,
                'alpha_t':est2.tvalues.const,
                'alpha_p':est2.pvalues.const,
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }  

    def _lambda_alpha_beta(self, df_i):
        df_i = df_i.dropna()
        if sum(df_i.fund) == 0 or sum(df_i.benchmark) == 0:
            return {'alpha':np.Inf,'beta':np.Inf}
        else:
            std_x = df_i.benchmark.var()
            std_y = df_i.fund.var()
            if std_x  == 0:
                print('##########boom  x')
            if std_y == 0:
                print('##########boom  y')
            ploy_res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            beta = ploy_res[0]
            alpha = ploy_res[1] * self._TRADING_DAYS_PER_YEAR
            return {'alpha':alpha,'beta':beta}

    def _get_benchmark_return(self) -> pd.DataFrame:
        benchmark_list: Dict[str, float] = {}
        fund_benchmark_df: pd.DataFrame = self._basic_api.get_fund_benchmark().drop(columns='_update_time')
        # 遍历每一只基金的benchmark进行处理
        for row in fund_benchmark_df.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index == '1':
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index in ('ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'):
                    #elif str(index_raw).startswith('RA000'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index].copy()
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                self.log.append((row.fund_id, index))
                                break
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index].copy()
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                self.log.append((row.fund_id, index))
                                break
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
                else:
                    if values or cons:
                        the_sum: float = sum(values)
                        if cons:
                            the_sum += np.log(math.pow(1 + cons, 1 / self._TRADING_DAYS_PER_YEAR))
                        benchmark_list[row.fund_id] = the_sum

        return pd.DataFrame.from_dict(benchmark_list)

    def get_continue_stats_tool(self):
        type_list = self._fund_list.wind_class_2.unique().tolist()
        self.rank_dic = {}
        for t_i in type_list:
            _fund_list_type = self._fund_list[self._fund_list['wind_class_2'] == t_i].fund_id.tolist()
            fund_ret_columns = self._fund_ret.columns.tolist()
            _fl = [ i for i in fund_ret_columns if i in _fund_list_type]
            resample_ret = self.df_resample(self._fund_ret[_fl])
            rank_df = resample_ret.rank(pct=True,axis=1)
            self.rank_dic[t_i] = rank_df.copy()

    def get_scale(self, end_date, fund_id):
        try:
            scale = self._fund_size.loc[self.end_date,fund_id]
        except:
            scale = np.nan
        return scale

    def calculate_item(self, fund_id):
        wind_2_class = self._fund_list[self._fund_list['fund_id'] == fund_id].wind_class_2.values[0]

        sr = self._fund_nav[fund_id].dropna().tail(self._TRADING_DAYS_PER_YEAR)
        df_i = self._fund_ret[[fund_id]].rename(columns={fund_id:'fund'}).join(self._benchmark_ret[[fund_id]].rename(columns={fund_id:'benchmark'})).tail(self._TRADING_DAYS_PER_YEAR)
        _sr = self._fund_nav[fund_id]
        l = _sr.shape[0]
        _res = []
        for i in range(1,19,1):
            bloc = -l+20*(i-1)
            eloc = int(self._TRADING_DAYS_PER_YEAR/2)+-l+20*i
            _res.append(self._sharpe(_sr.iloc[bloc:eloc]))
        _res = [i for i in _res if i == i and i != np.inf and i != -np.inf]
        if len(_res) < 6:
            continue_regress_v = np.nan
            continue_regress_t = np.nan
        else:
            mod = AutoReg(_res, 1)
            res = mod.fit()
            continue_regress_v = res.params[0]
            continue_regress_t = res.tvalues[0]

        _rank_df = self.rank_dic[wind_2_class]
        res = []
        last = ''
        for idx,i in enumerate(_rank_df[fund_id]):
            now = 'W' if i > 0.5 else 'L'
            if idx != 0:
                res.append(last+now)
            last = now
        ww = max(res.count('WW'),1)
        ll = max(res.count('LL'),1)
        wl = max(res.count('WL'),1)
        lw = max(res.count('LW'),1)
        continue_stats_v = ww*ll/wl/lw

        res_i = self._lambda_alpha_beta(df_i)
        alpha = res_i['alpha']
        beta = res_i['beta']
        mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
        mdd = 1 - mdd_part.min()
        annual_vol = (sr / sr.shift(1)).std(ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        annual_ret = np.exp(np.log(sr.iloc[-1] / sr.fillna(method='bfill').iloc[0]) / sr.shape[0] * self._TRADING_DAYS_PER_YEAR) - 1
        sharpe = (annual_ret - self._RISK_FEE_RATE) / annual_vol
        treynor = (self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id].mean() - self._RISK_FEE_RATE_PER_DAY)/beta
        calma_ratio = annual_ret / mdd if mdd != 0 else np.nan
        track_err = (df_i['fund'].tail(self._TRADING_DAYS_PER_YEAR) - df_i.tail(self._TRADING_DAYS_PER_YEAR)['benchmark']).std(ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        info_ratio = alpha / track_err
        scale = self.get_scale(self.end_date,fund_id)
        stock_tm_res = self._lambda_tm(self._benchmark_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id], self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR))
        stutzer = self.stutzer_index(self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id] - self._benchmark_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id])['stutzer_index']
        
        stock_tm_alpha = stock_tm_res['alpha']
        stock_tm_alpha_t = stock_tm_res['alpha_t']
        stock_tm_alpha_p = stock_tm_res['alpha_p']
        stock_tm_beta = stock_tm_res['beta']
        stock_tm_beta_t = stock_tm_res['beta_t']
        stock_tm_beta_p = stock_tm_res['beta_p']

        stock_hm_res = self._lambda_hm(self._benchmark_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id], self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR))
        stock_hm_alpha = stock_hm_res['alpha']
        stock_hm_alpha_t = stock_hm_res['alpha_t']
        stock_hm_alpha_p = stock_hm_res['alpha_p']
        stock_hm_beta = stock_hm_res['beta']
        stock_hm_beta_t = stock_hm_res['beta_t']
        stock_hm_beta_p = stock_hm_res['beta_p']

        stock_cl_res = self._lambda_cl(self._benchmark_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id], self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR))
        stock_cl_alpha = stock_cl_res['alpha']
        stock_cl_alpha_t = stock_cl_res['alpha_t']
        stock_cl_alpha_p = stock_cl_res['alpha_p']
        stock_cl_beta = stock_cl_res['beta']
        stock_cl_beta_t = stock_cl_res['beta_t']
        stock_cl_beta_p = stock_cl_res['beta_p']
        
        return {
            'fund_id':fund_id,
            'datetime':self.end_date,
            'alpha':alpha,
            'beta':beta,
            'annual_vol':annual_vol,
            'annual_ret':annual_ret,
            'track_err':track_err,
            
            'continue_regress_v':continue_regress_v,
            'continue_regress_t':continue_regress_t,
            'continue_stats_v':continue_stats_v,

            'stock_cl_alpha':stock_cl_alpha,
            'stock_cl_alpha_t':stock_cl_alpha_t,
            'stock_cl_alpha_p':stock_cl_alpha_p,
            'stock_cl_beta':stock_cl_beta,
            'stock_cl_beta_t':stock_cl_beta_t,
            'stock_cl_beta_p':stock_cl_beta_p,

            'stock_tm_alpha':stock_tm_alpha,
            'stock_tm_alpha_t':stock_tm_alpha_t,
            'stock_tm_alpha_p':stock_tm_alpha_p,
            'stock_tm_beta':stock_tm_beta,
            'stock_tm_beta_t':stock_tm_beta_t,
            'stock_tm_beta_p':stock_tm_beta_p,

            'stock_hm_alpha':stock_hm_alpha,
            'stock_hm_alpha_t':stock_hm_alpha_t,
            'stock_hm_alpha_p':stock_hm_alpha_p,
            'stock_hm_beta':stock_hm_beta,
            'stock_hm_beta_t':stock_hm_beta_t,
            'stock_hm_beta_p':stock_hm_beta_p,
            'info_ratio':info_ratio,
            'treynor':treynor,
            'sharpe':sharpe,
            'calma_ratio':calma_ratio,
            'stutzer':stutzer,
            'scale':scale,
        }        

    def _sharpe(self, sr):
        annual_vol = (sr / sr.shift(1)).std(ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        annual_ret = np.exp(np.log(sr.iloc[-1] / sr.iloc[0]) / sr.shape[0] * self._TRADING_DAYS_PER_YEAR) - 1
        sharpe = (annual_ret - self._RISK_FEE_RATE) / annual_vol
        return sharpe

    def calculate(self):
        self.res = []
        fund_list = self._fund_list[['fund_id','wind_class_1','wind_class_2']]
        _fund_list = self._fund_nav.columns.tolist()
        _fund_list_benchmark = self._benchmark_ret.columns.tolist()

        for i in fund_list.itertuples():
            if i.fund_id not in _fund_list or i.fund_id not in _fund_list_benchmark:
                continue
            self.res.append(self.calculate_item(i.fund_id))


import datetime
import pandas as pd


class DataManager:

    def __init__(self, start_time, end_time):
        # start_time: None, datetime or '20100101'
        # end_time:   None, datetime or '20100101'
        self.start_time = start_time if isinstance(start_time, datetime.datetime) else datetime.datetime.strptime(start_time,'%Y%m%d')
        self.end_time = end_time if isinstance(end_time, datetime.datetime) else datetime.datetime.strptime(end_time,'%Y%m%d')
        if self.end_time.date() == datetime.datetime.now().date():
            self.end_time = self.end_time - datetime.timedelta(days=1)

    def init(self):
        pass

    def terminate(self):
        pass

    @property
    def start_date(self):
        return self.start_time.date()

    @property
    def end_date(self):
        return self.end_time.date()
    
    


import pandas as pd
import traceback
from ...wrapper.mysql import DerivedDatabaseConnector, BasicDatabaseConnector, ViewDatabaseConnector
from ...view.view_models import IndexDailyCollection
from ...view.basic_models import IndexInfo, IndexPrice, IndexComponent
from ...view.derived_models import IndexValuation, IndexReturn, IndexVolatility

class IndexDailyCollectionProcessor(object):
    def get_index_volatility(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexVolatility).order_by(IndexVolatility.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexVolatility.index_id,
                    IndexVolatility.datetime,
                    IndexVolatility.w1_vol,
                    IndexVolatility.m1_vol,
                    IndexVolatility.m3_vol,
                    IndexVolatility.m6_vol,
                    IndexVolatility.y1_vol,
                    IndexVolatility.y3_vol,
                    IndexVolatility.y5_vol,
                    IndexVolatility.y10_vol,
                    IndexVolatility.this_y_vol,
                    IndexVolatility.cumulative_vol,
                ).filter(IndexVolatility.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'datetime': 'vol_datetime'})
                df = df.set_index('index_id')
                return df
            except Exception as e:
                print('Failed get_index_volatility <err_msg> {}'.format(e))

    # TODO: Use basic.index_valuation_develop table
    def get_index_valuation(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexValuation).order_by(IndexValuation.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexValuation.index_id,
                    IndexValuation.pb_mrq,
                    IndexValuation.pe_ttm,
                    IndexValuation.peg_ttm,
                    IndexValuation.roe,
                    IndexValuation.dy,
                    IndexValuation.pe_pct,
                    IndexValuation.pb_pct,
                    IndexValuation.ps_pct,
                    IndexValuation.val_score,
                    IndexValuation.datetime,
                    IndexValuation.est_peg,
                ).filter(IndexValuation.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_valuation <err_msg> {}'.format(e))

    def get_index_return(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexReturn).order_by(IndexReturn.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexReturn.index_id,
                    IndexReturn.datetime,
                    IndexReturn.w1_ret,
                    IndexReturn.m1_ret,
                    IndexReturn.m3_ret,
                    IndexReturn.m6_ret,
                    IndexReturn.y1_ret,
                    IndexReturn.y3_ret,
                    IndexReturn.y5_ret,
                    IndexReturn.y10_ret,
                    IndexReturn.this_y_ret,
                    IndexReturn.cumulative_ret,
                ).filter(IndexReturn.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'datetime': 'ret_datetime'})
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_return<err_msg> {}'.format(e))

    def get_index_info(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexInfo.index_id,
                    IndexInfo.em_id,
                    IndexInfo.industry_tag,
                    IndexInfo.tag_method,
                    IndexInfo.desc_name,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_info<err_msg> {}'.format(e))

    def get_index_component(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexComponent.index_id,
                    IndexComponent.id_cat,
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df['id_cat'] = df['id_cat'].apply(lambda x: x.name)
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_info<err_msg> {}'.format(e))

    def get_index_price(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(IndexPrice).order_by(IndexPrice.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    IndexPrice.index_id,
                    IndexPrice.datetime,
                    IndexPrice.volume,
                    IndexPrice.low,
                    IndexPrice.close,
                    IndexPrice.high,
                    IndexPrice.open,
                    IndexPrice.ret,
                    IndexPrice.total_turnover
                ).filter(IndexPrice.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'datetime': 'price_datetime'})
                df = df.set_index('index_id')
                return df

            except Exception as e:
                print('Failed get_index_price<err_msg> {}'.format(e))

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def collection_daily_index(self):
        try:
            print('1、 load data...')
            info = self.get_index_info()
            vol = self.get_index_volatility()
            ret = self.get_index_return()
            val = self.get_index_valuation()
            price = self.get_index_price()
            component = self.get_index_component()

            print('2、 concat data...')
            df = pd.concat([info, vol, ret, val, price, component], axis=1, sort=False)
            df.index.name = 'index_id'
            df = df.reset_index()

            print('3、 special option...')

            df['order_book_id'] = df['em_id'].apply(lambda x: x.split('.')[0] if (x and x != 'not_available') else None)
            df = df.drop(['em_id'], axis=1)
            df['industry_tag'] = df['industry_tag'].apply(lambda x: x if x != 'not_available' else None)

            self.append_data(IndexDailyCollection.__tablename__, df)

            print(df)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.collection_daily_index():
            failed_tasks.append('collection_daily_index')
        return failed_tasks


if __name__ == '__main__':
    IndexDailyCollectionProcessor().collection_daily_index()



import os
import json
import platform
from .singleton import Singleton

DEFAULT_CONFIG_PATH = '/shared/surfing/etc/config.json'
DEFAULT_LOG_DIR = '/shared/surfing/log/'
DEFAULT_TMP_DIR = '/shared/surfing/tmp/'

class SysSettings(object):

    def __init__(self, sys_json):
        self.log_dir = sys_json.get('log_dir', DEFAULT_LOG_DIR)
        self.tmp_dir = sys_json.get('tmp_dir', DEFAULT_TMP_DIR)

        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        if not os.path.exists(self.tmp_dir):
            os.makedirs(self.tmp_dir)
    
    def __str__(self):
        return f'log_dir: {self.log_dir}\ntmp_dir: {self.tmp_dir}'

class DbSettings(object):
    '''
    Parse database connection settings in config file
    '''
    def __init__(self, db_json, db_name=None, default_settings=None):
        self.host = db_json.get('host', '' if not default_settings else default_settings.host)
        self.port = db_json.get('port', 0 if not default_settings else default_settings.port)
        self.username = db_json.get('username', '' if not default_settings else default_settings.username)
        self.password = db_json.get('password', '' if not default_settings else default_settings.password)
        self.db_name = db_name

    def to_conn_str(self):
        return 'mysql+pymysql://{}:{}@{}:{}/{}'.format(
            self.username, self.password, self.host, self.port, self.db_name
        )

    def __str__(self):
        return 'mysql+pymysql://{}:********@{}:{}/{}'.format(
            self.username, self.host, self.port, self.db_name
        )

class CassandraSettings(object):

    def __init__(self, conf_json):
        self.hosts = conf_json.get('hosts', '')
        self.port = conf_json.get('port', '')
        self.username = conf_json.get('username', '')
        self.password = conf_json.get('password', '')
        self.default_key_space = conf_json.get('default_key_space', None)

    def to_dict(self):
        return {
            "hosts": self.hosts,
            "port": self.port,
            "username": self.username,
            "password": self.password,
        }

    def __str__(self):
        return 'hosts:{} ** port:{} ** username:{}** password:{}'.format(
            self.hosts, self.port, self.username, self.password,
        )

class SurfingConfigurator(metaclass=Singleton):

    def __init__(self, config_path=DEFAULT_CONFIG_PATH):
        with open(config_path) as f:
            self.config = json.load(f)

    def get_sys_settings(self):
        assert 'sys' in self.config, 'could not find sys in config'
        settings = SysSettings(self.config['sys'])
        return settings

    def get_db_settings(self, db_name=None):
        assert 'db' in self.config, 'could not find db in config'
        db_config = self.config['db']
        default_settings = None if db_name == '_default' else self.get_db_settings('_default')
        db_settings = DbSettings(db_config.get(db_name, {}), db_name, default_settings)
        return db_settings

    def get_license_settings(self, key=None):
        assert 'license' in self.config, 'could not find license section in config'
        settings = self.config['license']
        assert key in settings, f'could not find {key} section in license config'
        return settings[key]

    def get_wechat_webhook_settings(self, key=None):
        assert 'wechat_webhook' in self.config, 'could not find wechat_webhook section in config'
        settings = self.config['wechat_webhook']
        return settings

    def get_cassandra_setting(self):
        assert 'surfing_cassandra' in self.config, 'could not find surfing_cassandra section in config'
        return CassandraSettings(self.config['surfing_cassandra'])



from .backtest_engine import BacktestEngine
from .strategy_base import StrategyBase, Direction
from ..structure import AssetType, PositionInfo

class StockDailyStrategy(StrategyBase):

    def __init__(self,strategy_name=None,pool_name=None,factor_name=None,data_path=None,longshort=None,top=None,period=None):

        self.factor_name    = factor_name
        self.longshort      = longshort
        self.top            = top
        self.sell_list      = []
        self.wait_sell_list = [] 
        self.period         = period
        self.cash           = 0
        # input fee here to pre calculate amount of cash used in buying stock
        self.sell_fee       = PositionInfo.FEE_RATE['STOCK']['SLIPPAGE']+PositionInfo.FEE_RATE['STOCK']['SELLFEE']
        self.buy_fee        = PositionInfo.FEE_RATE['STOCK']['SLIPPAGE']+PositionInfo.FEE_RATE['STOCK']['BUYFEE']
        self.cash_spliter   = 1
        StrategyBase.__init__(self, strategy_name)

    def init(self):
        self.pay_count = 5
        self.stop_count = 10
        self.index = 0

    def on_day(self, day_data):
        # strategy main function, daily buy & sell
        self.cash = self.context._order_manager.positions[self._name].cash.position
        self.build_cash_spliter(day_data)        
        sell_day = self.get_sell_day(day_data['time'], self.period)
        if sell_day:
            self.sell_list = self.get_trade_buy(sell_day)
        self.build_long()
        self.build_trade_list(day_data)

    def build_cash_spliter(self, day_data):
        idx = self.context._data_manager.wrapper.date_list.index(day_data['time'])
        if idx < (self.period-1):
            self.cash_spliter = self.period - idx
        else:
            self.cash_spliter = 1

    def build_long(self):
        if self.longshort[0] == 0:
            self.long_list = None
        else:
            ticker_list = [k for k, v in sorted(self.context._data_manager.factor_today.items(), key=lambda item: item[1]) if k not in self.context._data_manager.rm_l_today]
            if self.longshort[0] == 1:
                self.long_list = ticker_list[self.top * -1:]
            elif self.longshort[0] == -1:
                self.long_list = ticker_list[: self.top ]
            
    def build_short(self):
        #TODO such shorttop
        pass

    def build_trade_list(self, day_data):
        # if tick should be sell today but suspended, put it into wait_sell_list
        self.order_list = []
        if self.sell_list:
            sell_list = self.sell_list + self.wait_sell_list
            self.wait_sell_list = []
            for sell_i in sell_list:
                if sell_i.ticker in self.context._data_manager.suspended_today:
                    self.wait_sell_list.append(sell_i)
                    continue
                od = self.build_order(ticker=sell_i.ticker, direction=Direction.SELL, volume=sell_i.volume, price=day_data[sell_i.ticker], asset_type=AssetType.STOCK)
                self.order_list.append(od)
                self.cash += sell_i.volume * day_data[sell_i.ticker] * (1 - self.sell_fee )
        
        for long_i in self.long_list:
            price = day_data[long_i]
            od = self.build_order(ticker=long_i, direction=Direction.BUY, volume=self.cash/self.cash_spliter/price/self.top/(1+self.buy_fee), price=price,asset_type=AssetType.STOCK)
            self.order_list.append(od)

    def on_trade(self, trade):
        pass

    def on_order(self, order_id, order_status):
        pass





from sqlalchemy import CHAR, Column, Integer, Index, BOOLEAN, text, TEXT, Enum
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT, DATETIME, MEDIUMTEXT

from ...constant import SectorType, IndexPriceSource


class Base():
    _update_time = Column('_update_time', DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间


# make this column at the end of every derived table
Base._update_time._creation_order = 9999
Base = declarative_base(cls=Base)


class SectorInfo(Base):
    '''板块信息表'''

    __tablename__ = 'sector_info'

    sector_id = Column(CHAR(64), primary_key=True)  # 板块ID
    sector_type = Column(Enum(SectorType), nullable=False)  # 板块类型 industry-行业 topic-主题
    sector_desc = Column(TEXT, nullable=False)  # 板块描述
    sector_name = Column(CHAR(16), nullable=False, unique=True)  # 板块名称
    main_index_id = Column(CHAR(16), nullable=False)  # 板块代表指数


class SectorFunds(Base):
    '''板块基金表'''

    __tablename__ = 'sector_funds'

    id = Column(Integer, primary_key=True)

    sector_id = Column(CHAR(64), nullable=False)  # 板块ID
    csi_index_id = Column(CHAR(16), nullable=False)  # Choice指数ID
    index_id = Column(CHAR(20), nullable=False)  # 指数ID
    em_fund_id = Column(CHAR(16))  # Choice基金ID
    fund_id = Column(CHAR(16))  # 基金ID


class IndexInfo(Base):
    '''指数信息表'''

    __tablename__ = 'index_info'
    index_id = Column(CHAR(20), primary_key=True)

    order_book_id = Column(CHAR(20))  # 米筐ID
    # web_id= Column(CHAR(20)) # 数据所在网页ID
    industry_tag = Column(CHAR(64))  # 行业标签
    tag_method = Column(CHAR(64))  # 估值评分采用方法
    desc_name = Column(CHAR(64), nullable=False, unique=True)  # 名称
    maker_name = Column(CHAR(32), nullable=False)  # 编制机构
    # publish_date = Column(DATE, nullable=False)  # 发布日期
    is_select = Column(BOOLEAN, nullable=False)  # 是否属于精选
    em_id = Column(CHAR(20))  # ChoiceID
    em_plate_id = Column(CHAR(20))  # Choice板块ID
    index_profile = Column(TEXT)  # 指数概况
    price_source = Column(Enum(IndexPriceSource), nullable=False)  # 指数价格来源标记


class IndexComponent(Base):

    __tablename__ = 'index_component'

    index_id = Column(CHAR(20), primary_key=True)
    datetime = Column(DATE, primary_key=True)  # 日期
    num = Column(Integer, nullable=False)  # 成分股数量
    id_cat = Column(Enum(SectorType), nullable=False)  # 所属类型 industry-行业 topic-主题
    sector = Column(TEXT, nullable=False)  # 所属板块
    top10 = Column(TEXT, nullable=False)  # 前10成分及权重
    related_funds = Column(TEXT)  # 相关产品
    all_constitu = Column(MEDIUMTEXT, nullable=False)  # 所有成分及权重(from Choice)


class StockInfo(Base):
    '''股票信息表'''

    __tablename__ = 'stock_info'
    stock_id = Column(CHAR(20), primary_key=True) # 股票ID
    rq_id = Column(CHAR(20)) # 米筐ID


class FundSize(Base):
    '''基金最新规模'''

    __tablename__ = 'fund_size'

    fund_id = Column(CHAR(10), primary_key=True) # 基金id
    latest_size = Column(DOUBLE(asdecimal=False)) # 最新规模


class FundStatusLatest(Base):
    '''
    基金状态表 包含基金申赎状态等
    '''

    __tablename__ = 'fund_status_latest'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金ID

    PURCHSTATUS = Column('purch_status', CHAR(16))  # 申购状态
    REDEMSTATUS = Column('redem_status', CHAR(16))  # 赎回状态
    LPLIMIT = Column('lp_limit', DOUBLE(asdecimal=False))  # 暂停大额申购上限
    OTCLPLIMIT = Column('otc_lp_limit', DOUBLE(asdecimal=False))  # 场外暂停大额申购上限
    TRADESTATUS = Column('trade_status', CHAR(16))  # 交易状态


class FundInfo(Base):
    '''基金信息表'''

    __tablename__ = 'fund_info'
    fund_id = Column(CHAR(10), primary_key=True) # 基金ID

    wind_id = Column(CHAR(20)) # Wind基金ID
    transition = Column(Integer) # 基金变更次数
    order_book_id = Column(CHAR(10)) # RiceQuant基金ID
    desc_name = Column(CHAR(64)) # 基金名称
    start_date = Column(DATE) # 成立日期
    end_date = Column(DATE) # 关闭日期
    wind_class_1 = Column(CHAR(64)) # Wind基金类型
    wind_class_2 = Column(CHAR(64)) # Wind基金二级类型
    manager_id = Column(TEXT) # 基金经理
    company_id = Column(CHAR(64)) # 基金公司
    benchmark = Column(CHAR(255)) # 业绩基准
    full_name = Column(CHAR(255)) # 基金全名
    currency = Column(CHAR(20)) # 币种
    base_fund_id = Column(CHAR(20)) # 分级基金基础基金代号
    is_structured = Column(TINYINT(1)) # 是否为分级基金
    is_open = Column(TINYINT(1)) # 是否为日常开放申赎的基金 1是 0否， 排除掉封闭基金和定期开放基金
    is_regular_open_ended = Column(TINYINT(1)) # 是否为定期开放式基金 1是 0否
    is_closed_ended= Column(TINYINT(1)) # 是否为封闭基金 1是 0 否
    structure_type = Column(TINYINT(1))  # 是否为分级子基金  0/nan非   1母   2分级A  3 分级B 4 其他
    is_etf = Column(TINYINT(1)) # 是否是etf  0 非  1 etf
    asset_type = Column(CHAR(32)) # 资产类别
    tt_purchase_fee = Column(DOUBLE(asdecimal=False)) # 天天基金优惠申购费
    manage_fee = Column(DOUBLE(asdecimal=False)) # 管理费
    trustee_fee = Column(DOUBLE(asdecimal=False)) # 托管费
    purchase_fee = Column(DOUBLE(asdecimal=False)) # 申购费
    redeem_fee = Column(DOUBLE(asdecimal=False)) # 赎回费
    note = Column(CHAR(64)) # 附加信息
    track_index = Column(CHAR(20)) # 跟踪指数
    benchmark_1 = Column(CHAR(255)) # 业绩基准标的指数简称第一名
    benchmark_2 = Column(CHAR(255)) # 业绩基准标的指数简称第二名
    index_id = Column(CHAR(20)) # 基于第一业绩标准所标注的指数ID
    is_c =  Column("is_c", BOOLEAN, nullable=False)  #是否是etf  0 非  1 c
    is_a = Column("is_a", BOOLEAN, nullable=False)  #是否是etf  0 非  1 a
    ac_filter = Column("ac_filter", BOOLEAN, nullable=False) #是否ac均有要排除的a 0 排除  1 保留
    is_selected_mmf = Column(TINYINT(1)) # 是否基金规模最大的前50只货币基金(10年前有两只存在)  0 非  1 是
    national_debt_extension = Column(TINYINT(1)) # 纯债tag  0 非  1 是
    index_id_new = Column(CHAR(20)) # 基于第一业绩标准所标注的指数ID新
    subscr_fee_detail = Column(TEXT) # 认购费率详情
    purchase_fee_detail = Column(TEXT) # 申购费率详情
    redeem_fee_detail = Column(TEXT) # 赎回费率详情
    service_fee = Column(TEXT) # 销售服务费率
    fund_manager = Column(TEXT)  # 基金经理(现任)
    pre_fund_manager = Column(TEXT)  # 基金经理(历任)


class FundBenchmark(Base):
    '''基金业绩比较基准表'''

    __tablename__ = 'fund_benchmark'

    em_id = Column(CHAR(16), primary_key=True)  # 基金代码
    fund_id = Column(CHAR(16))  # 内部基金代码
    # datetime = Column(DATE, primary_key=True)  # 日期
    index_text = Column(TEXT, nullable=False)  # 业绩比较基准
    benchmark = Column(TEXT, nullable=False)  # 解析出来的业绩比较基准
    benchmark_s = Column(TEXT, nullable=False)  # index转换后的业绩比较基准
    assets = Column(CHAR(32))  # 追踪的标的
    industry = Column(CHAR(64))  # 行业分类

    # __table_args__ = (
    #     Index('idx_em_fund_benchmark_datetime', 'datetime'),
    # )


class TradingDayList(Base):
    '''交易日列表'''

    __tablename__ = 'trading_day_list'
    datetime = Column(DATE, primary_key=True)


class FundNav(Base):
    '''基金净值表'''

    __tablename__ = 'fund_nav'
    fund_id = Column(CHAR(20), primary_key=True) # 合约代码
    datetime = Column(DATE, primary_key=True) # 日期

    unit_net_value = Column(DOUBLE(asdecimal=False)) # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False)) # 累计单位净值 基金公司公告发布的原始数据 单位净值 + 单位累计分红
    adjusted_net_value = Column(DOUBLE(asdecimal=False)) # 复权净值 考虑分红再投资后调整的单位净值
    change_rate = Column(DOUBLE(asdecimal=False)) # 涨跌幅
    daily_profit = Column(DOUBLE(asdecimal=False)) # 每万元收益（日结型货币基金专用）
    weekly_yield = Column(DOUBLE(asdecimal=False)) # 7日年化收益率（日结型货币基金专用）
    redeem_status = Column(Integer) # 赎回状态，开放 - Open, 暂停 - Suspended, 限制大额申赎 - Limited, 封闭期 - Close
    subscribe_status = Column(Integer) # 订阅状态，开放 - Open, 暂停 - Suspended, 限制大额申赎 - Limited, 封闭期 - Close
    fund_size = Column(DOUBLE(asdecimal=False)) # 基金规模

    __table_args__ = (
        Index('idx_fund_nav_datetime', 'datetime'),
    )


class IndexPrice(Base):
    '''指数价格表'''

    __tablename__ = 'index_price'
    index_id = Column(CHAR(20), primary_key=True) # 指数id
    datetime = Column(DATE, primary_key=True) # 日期

    volume = Column(DOUBLE(asdecimal=False)) # 交易量
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 成交额
    ret = Column(DOUBLE(asdecimal=False)) # 日收益

    __table_args__ = (
        Index('idx_index_price_datetime', 'datetime'),
    )

class StockPrice(Base):
    '''股票价格表'''

    __tablename__ = 'stock_price'
    stock_id = Column(CHAR(20), primary_key=True) # 指数id
    datetime = Column(DATE, primary_key=True) # 日期

    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    limit_up = Column(DOUBLE(asdecimal=False)) # 涨停价
    limit_down = Column(DOUBLE(asdecimal=False)) # 跌停价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    volume = Column(DOUBLE(asdecimal=False)) # 交易量
    num_trades = Column(DOUBLE(asdecimal=False)) # 交易笔数
    adj_close = Column(DOUBLE(asdecimal=False)) # 后复权价格
    post_adj_factor = Column(DOUBLE(asdecimal=False)) # 后复权因子

    __table_args__ = (
        Index('idx_stock_price_datetime', 'datetime'),
    )


class FundRet(Base):
    '''基金历史收益'''

    __tablename__ = 'fund_ret'

    fund_id = Column(CHAR(10), primary_key=True) # 原始基金ID
    datetime = Column(DATE, primary_key=True) # 日期

    w1_ret = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_ret = Column(DOUBLE(asdecimal=False)) # 近一月收益率
    m3_ret = Column(DOUBLE(asdecimal=False)) # 近一季度收益率
    m6_ret = Column(DOUBLE(asdecimal=False)) # 近半年收益率
    y1_ret = Column(DOUBLE(asdecimal=False)) # 近一年收益率
    y3_ret = Column(DOUBLE(asdecimal=False)) # 近三年收益率
    y5_ret = Column(DOUBLE(asdecimal=False)) # 近五年收益率
    to_date_ret = Column(DOUBLE(asdecimal=False)) # 成立至今收益率
    mdd = Column(DOUBLE(asdecimal=False)) # 最大回撤（成立以来）
    annual_ret = Column(DOUBLE(asdecimal=False)) # 年化收益（成立以来）
    avg_size = Column(DOUBLE(asdecimal=False)) # 平均规模（成立以来）
    sharpe_ratio = Column(DOUBLE(asdecimal=False)) # 夏普率（成立以来）
    vol = Column(DOUBLE(asdecimal=False)) # 波动率（成立以来）
    info_ratio = Column(DOUBLE(asdecimal=False)) # 信息比率（成立以来）
    recent_y_ret = Column(DOUBLE(asdecimal=False)) # 年初至今收益率

class FundRatingLatest(Base):
    '''基金最新评级'''

    __tablename__ = 'fund_rating_latest'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金id
    zs = Column(DOUBLE(asdecimal=False))  # 招商评级
    sh3 = Column(DOUBLE(asdecimal=False))  # 上海证券评级三年期
    sh5 = Column(DOUBLE(asdecimal=False))  # 上海证券评级五年期
    jajx = Column(DOUBLE(asdecimal=False))  # 济安金信评级
    update_time = Column(DATE)  # 更新日期

    __table_args__ = (
        Index('idx_fund_rating_latest_datetime', 'update_time'),
    )


class StyleAnalysisStockFactor(Base):
    '''风格分析股票因子'''

    __tablename__ = 'style_analysis_stock_factor'

    stock_id = Column(CHAR(10), primary_key=True)  # EM股票ID
    datetime = Column(DATE, primary_key=True)  # 日期
    rate_of_return = Column(DOUBLE(asdecimal=False), nullable=False)  # 收益率
    latest_size = Column(DOUBLE(asdecimal=False))  # 规模
    bp = Column(DOUBLE(asdecimal=False))  # 价值
    short_term_momentum = Column(DOUBLE(asdecimal=False))  # 短期动量
    long_term_momentum = Column(DOUBLE(asdecimal=False))  # 长期动量
    high_low = Column(DOUBLE(asdecimal=False))  # 波动率

    __table_args__ = (
        Index('idx_style_analysis_stock_factor_datetime', 'datetime'),
    )

class Fund_size_and_hold_rate(Base):
    '''基金规模和持有人比例'''

    __tablename__ = 'fund_size_and_hold_rate'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金id
    datetime = Column(DATE, primary_key=True)  # 日期
    size = Column(DOUBLE(asdecimal=False)) # 最新规模
    personal_holds = Column(DOUBLE(asdecimal=False)) # 个人持有比例 单位百分比
    institution_holds = Column(DOUBLE(asdecimal=False)) # 机构持有比例 单位百分比

class FundRate(Base):
    '''基金评级'''

    __tablename__ = 'fund_rate'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金id
    datetime = Column(DATE, primary_key=True)  # 日期
    mkt_3y_comp = Column(CHAR(5)) # 市场综合3年评级
    mornstar_3y =  Column(CHAR(5)) # 晨星3年评级
    mornstar_5y =  Column(CHAR(5)) # 晨星5年评级
    merchant_3y = Column(CHAR(5)) # 招商3年评级
    tx_comp = Column(CHAR(5)) # 天相投顾综合评级
    sh_3y_comp = Column(CHAR(5)) #上海证券3年评级(综合评级)
    sh_3y_stock = Column(CHAR(5)) #上海证券3年评级(选股能力)
    sh_3y_timeret = Column(CHAR(5)) #上海证券3年评级(择时能力)
    sh_3y_sharpe = Column(CHAR(5)) #上海证券3年评级(夏普比率)
    sh_5y_comp = Column(CHAR(5)) #上海证券5年评级(综合评级)
    sh_5y_stock = Column(CHAR(5)) #上海证券5年评级(选股能力)
    sh_5y_timeret = Column(CHAR(5)) #上海证券5年评级(择时能力)
    sh_5y_sharpe = Column(CHAR(5)) #上海证券5年评级(夏普比率)
    jian_comp = Column(CHAR(5)) #济安金信基金评级(综合评级)
    jian_earn = Column(CHAR(5)) #济安金信基金评级(盈利能力)
    jian_stable = Column(CHAR(5)) #济安金信基金评级(业绩稳定性)
    jian_risk = Column(CHAR(5)) #济安金信基金评级(抗风险能力)
    jian_stock = Column(CHAR(5)) #济安金信基金评级(选股能力)
    jian_timeret = Column(CHAR(5)) #济安金信基金评级(择时能力)
    jian_track = Column(CHAR(5)) #济安金信基金评级(基准跟踪能力)
    jian_alpha = Column(CHAR(5)) #济安金信基金评级(超额收益能力)
    jian_fee = Column(CHAR(5)) #济安金信基金评级(整体费用)

class FundHoldAsset(Base):
    '''基金持仓资产比例'''

    __tablename__ = 'fund_hold_asset'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金id
    datetime = Column(DATE, primary_key=True)  # 日期
    stock_nav_ratio = Column(DOUBLE(asdecimal=False)) # 股票市值占基金资产净值比
    bond_nav_ratio = Column(DOUBLE(asdecimal=False)) # 债券市值占基金资产净值比
    cash_nav_ratio = Column(DOUBLE(asdecimal=False)) # 银行存款占基金资产净值比
    other_nav_ratio = Column(DOUBLE(asdecimal=False)) # 其他资产占基金资产净值比

class FundHoldIndustry(Base):
    '''基金持仓前三行业'''

    __tablename__ = 'fund_hold_industry'

    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    rank1_indname = Column(CHAR(50)) #第一排名行业名称
    rank1_indweight = Column(DOUBLE(asdecimal=False)) #第一排名行业市值占基金资产总值比
    rank2_indname = Column(CHAR(50)) #第二排名行业名称
    rank2_indweight = Column(DOUBLE(asdecimal=False)) #第二排名行业市值占基金资产总值比
    rank3_indname = Column(CHAR(50)) #第三排名行业名称
    rank3_indweight = Column(DOUBLE(asdecimal=False)) #第三排名行业市值占基金资产总值比

class FundHoldStock(Base):
    '''基金持仓前十股票'''

    __tablename__ = 'fund_hold_stock'

    fund_id = Column(CHAR(10), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    rank1_stock = Column(CHAR(20)) #股票名
    rank1_stock_code = Column(CHAR(16))  # 股票ID
    rank1_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank1_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank2_stock = Column(CHAR(20)) #股票名
    rank2_stock_code = Column(CHAR(16))  # 股票ID
    rank2_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank2_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank3_stock = Column(CHAR(20)) #股票名
    rank3_stock_code = Column(CHAR(16))  # 股票ID
    rank3_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank3_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank4_stock = Column(CHAR(20)) #股票名
    rank4_stock_code = Column(CHAR(16))  # 股票ID
    rank4_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank4_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank5_stock = Column(CHAR(20)) #股票名
    rank5_stock_code = Column(CHAR(16))  # 股票ID
    rank5_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank5_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank6_stock = Column(CHAR(20)) #股票名
    rank6_stock_code = Column(CHAR(16))  # 股票ID
    rank6_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank6_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank7_stock = Column(CHAR(20)) #股票名
    rank7_stock_code = Column(CHAR(16))  # 股票ID
    rank7_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank7_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank8_stock = Column(CHAR(20)) #股票名
    rank8_stock_code = Column(CHAR(16))  # 股票ID
    rank8_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank8_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank9_stock = Column(CHAR(20)) #股票名
    rank9_stock_code = Column(CHAR(16))  # 股票ID
    rank9_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank9_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank10_stock = Column(CHAR(20)) #股票名
    rank10_stock_code = Column(CHAR(16))  # 股票ID
    rank10_stockval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank10_stockweight = Column(DOUBLE(asdecimal=False)) #持仓权重

class FundHoldBond(Base):
    '''东财基金重仓债'''

    __tablename__ = 'fund_hold_bond'
    fund_id = Column(CHAR(20), primary_key=True) # 基金ID
    datetime = Column(DATE, primary_key=True) # 日期
    rank1_bond = Column(CHAR(30)) #债券名
    rank1_bond_code =  Column(CHAR(15)) # 债券id
    rank1_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank1_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank2_bond = Column(CHAR(30)) #债券名
    rank2_bond_code =  Column(CHAR(15)) # 债券id
    rank2_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank2_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank3_bond = Column(CHAR(30)) #债券名
    rank3_bond_code =  Column(CHAR(15)) # 债券id
    rank3_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank3_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank4_bond = Column(CHAR(30)) #债券名
    rank4_bond_code =  Column(CHAR(15)) # 债券id
    rank4_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank4_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank5_bond = Column(CHAR(30)) #债券名
    rank5_bond_code =  Column(CHAR(15)) # 债券id
    rank5_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank5_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank6_bond = Column(CHAR(30)) #债券名
    rank6_bond_code =  Column(CHAR(15)) # 债券id
    rank6_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank6_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank7_bond = Column(CHAR(30)) #债券名
    rank7_bond_code =  Column(CHAR(15)) # 债券id
    rank7_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank7_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank8_bond = Column(CHAR(30)) #债券名
    rank8_bond_code =  Column(CHAR(15)) # 债券id
    rank8_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank8_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank9_bond = Column(CHAR(30)) #债券名
    rank9_bond_code =  Column(CHAR(15)) # 债券id
    rank9_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank9_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重
    rank10_bond = Column(CHAR(30)) #债券名
    rank10_bond_code =  Column(CHAR(15)) # 债券id
    rank10_bondval = Column(DOUBLE(asdecimal=False)) #持仓市值
    rank10_bondweight = Column(DOUBLE(asdecimal=False)) #持仓权重


import boto3
import json
import decimal
from pprint import pprint

TABLE_NAME = 'backtest_report'
backtest_Table = boto3.resource('dynamodb').Table(TABLE_NAME)

def decimal_form(number):
    return decimal.Decimal(number).quantize(decimal.Decimal("0.0000"))

def float_form(number):
    return float(number)

def recursion_decimal_dict(dic):
    for key in dic:
        if isinstance(dic[key], dict):
            recursion_decimal_dict(dic[key])
        else:
            if isinstance(dic[key], list):
                if isinstance(dic[key][1], float):
                    dic[key] = [ decimal_form(_) for _ in dic[key] ]
            elif isinstance(dic[key], float):
                dic[key] = decimal_form(dic[key])
            else:
                pass
    return dic

def backtest_result_to_dynamodb(data):
    data = recursion_decimal_dict(data)
    backtest_Table.put_item(Item=data, ReturnConsumedCapacity='TOTAL')




import pandas as pd
import numpy as np
import os
import datetime
import traceback
import requests
import urllib3
from ...wrapper.mysql import RawDatabaseConnector
from .raw_data_helper import RawDataHelper


class WebRawDataDownloader(object):

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def read_csv(self, csv_path):
        return pd.read_csv(csv_path, index_col=1)

    def fund_fee(self):
        pass

    def wind_fund_info(self, csv_path):
        # Update manually
        # Default file in company wechat is 'fundlist_wind_20202015.xlsx'
        df = pd.read_excel(csv_path)
        df.columns = ['wind_id', 'desc_name', 'full_name', 'start_date', 'end_date', 'benchmark', 'wind_class_1',
            'wind_class_2', 'currency', 'base_fund_id', 'is_structured', 'is_open', 'manager_id', 'company_id']
        df['start_date'] = df['start_date'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d'))
        df['end_date'] = df['end_date'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d')
            if isinstance(x, str) else datetime.datetime.strptime('2040-12-31', '%Y-%m-%d'))
        df['is_structured'] = df['is_structured'].map(lambda x: False if x == '否' else True)
        df['is_open'] = df['is_open'].map(lambda x: False if x == '否' else True)
        self._data_helper._upload_raw(df, 'wind_fund_info')

    def cm_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 汇率数据.xls
            # df = pd.read_excel(csv_path)
            # df = df[['日期','美元中间价','欧元中间价','日元中间价','美元CFETS','欧元CFETS','日元CFETS']]
            # df.columns = ['datetime','usd_central_parity_rate','eur_central_parity_rate','jpy_central_parity_rate',
            #               'usd_cfets','eur_cfets','jpy_cfets']

            # auto update
            # data from http://www.chinamoney.com.cn/chinese/bkccpr
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
            end_date = datetime.datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
            url = ('http://www.chinamoney.com.cn/dqs/rest/cm-u-bk-ccpr/CcprHisExcelNew'
                f'?startDate={start_date}&endDate={end_date}&currency=USD/CNY,EUR/CNY,100JPY/CNY')
            headers = {'User-Agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 '
                                      '(KHTML, like Gecko) Version/13.0.5 Safari/605.1.15')}
            filename = ''
            try:
                r = requests.get(url, headers=headers)
                filename = os.path.join('/tmp', 'cm.xlsx')
                with open(filename, 'wb') as f:
                    f.write(r.content)
            except Exception:
                pass
            df = pd.read_excel(filename)
            # Drop last 2 lines in excel, which are
            #   数据来源：	中国货币网
            #   www.chinamoney.com.cn
            df.drop(df.index[-2:], inplace=True)
            df = df[['日期', 'USD/CNY', 'EUR/CNY', '100JPY/CNY']]
            df.columns = ['datetime', 'usd_central_parity_rate', 'eur_central_parity_rate', 'jpy_central_parity_rate']
            df['jpy_central_parity_rate'] = df['jpy_central_parity_rate'] / 100
            df = df.sort_values('datetime')
            self._data_helper._upload_raw(df, 'cm_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def cxindex_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 中证信用债指数历史数据.xls
            # df = pd.read_excel(csv_path)
            # df.columns = ['index_id', 'symbol', 'datetime', 'open', 'high', 'low', 'close', 'total_turnover', 'volume']
            # df['ret'] = df['close'] / df['close'].shift(1) - 1
            # df['index_id'] = 'credit_debt'
            # df = df.drop(['symbol'], axis=1)[:-2]

            # auto update
            # data from http://www.csindex.com.cn/zh-CN/indices/index-detail/H11073
            # File download url
            url = 'http://www.csindex.com.cn/uploads/file/autofile/perf/H11073perf.xls'
            filename = ''
            try:
                r = requests.get(url)
                filename = os.path.join('/tmp', 'H11073perf.xls')
                with open(filename, 'wb') as f:
                    f.write(r.content)
            except:
                pass
            df = pd.read_excel(filename)
            df = df[['日期Date', '收盘Close', '涨跌幅(%)Change(%)', '成交量（万元）Volume(10 thousand CNY)',
                '成交金额（元）Turnover']]
            df.columns = ['datetime', 'close', 'ret', 'volume', 'total_turnover']
            df['volume'] = df['volume'].values * 10000
            df['ret'] = df['ret'] / 100
            df = df[(df['datetime']>=start_date) & (df['datetime']<=end_date)].sort_values('datetime')
            df['open'] = float('Nan')
            df['high'] = float('Nan')
            df['low'] = float('Nan')
            df['index_id'] = 'credit_debt'
            self._data_helper._upload_raw(df, 'cxindex_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def yahoo_index_price(self, start_date, end_date):
        try:
            # history
            # default file in company wechat is 大类资产指数据.xlsx
            # df = pd.read_excel(csv_path)
            # df['datetime'] = df['date']
            # for c in ['sp500', 'dax30', 'n225']:
            #     df_i = df[['datetime']].copy()
            #     df_i['close'] = df[c].copy()
            #     df_i['ret'] = df_i['close'] / df_i['close'].shift(1) - 1
            #     df_i['open'] = float('Nan')
            #     df_i['high'] = float('Nan')
            #     df_i['low'] = float('Nan')
            #     df_i['volume'] = float('Nan')
            #     df_i['total_turnover'] = float('Nan')
            #     df_i['index_id'] = c
            #     self._data_helper._upload_raw(df_i, 'yahoo_index_price')

            # auto update
            # data from
            # sp500 https://finance.yahoo.com/quote/%5EGSPC/history?p=%5EGSPC
            # n225 https://finance.yahoo.com/quote/%5EN225/history?p=%5EN225
            # dax30 https://finance.yahoo.com/quote/%5EGDAXI/history/
            # File download url:
            # https://query1.finance.yahoo.com/v7/finance/download/%5EGSPC?period1=1553667756&period2=1585290156&interval=1d&events=history
            # https://query1.finance.yahoo.com/v7/finance/download/%5EN225?period1=1553673330&period2=1585295730&interval=1d&events=history
            # https://query1.finance.yahoo.com/v7/finance/download/%5EGDAXI?period1=1553673346&period2=1585295746&interval=1d&events=history

            url_prefix = 'https://query1.finance.yahoo.com/v7/finance/download/'

            start_timestamp = str(int((datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=7)).timestamp()))
            end_timestamp = str(int((datetime.datetime.strptime(end_date, '%Y%m%d') + datetime.timedelta(days=1)).timestamp()))
            url_postfix = f'?period1={start_timestamp}&period2={end_timestamp}&interval=1d&events=history'
            url_type_list = ['%5EGSPC', '%5EN225', '%5EGDAXI']
            csv_list = ['^GSPC.csv', '^N225.csv', '^GDAXI.csv']
            index_list = ['sp500', 'n225', 'dax30']
            df_list = []
            for url_type, csv_i, index_i in zip(url_type_list, csv_list, index_list):
                url = url_prefix + url_type + url_postfix
                filename = ''
                try:
                    r = requests.get(url)
                    filename = os.path.join('/tmp', csv_i)
                    with open(filename, 'wb') as f:
                        f.write(r.content)
                except Exception:
                    pass
                df = pd.read_csv(filename)
                df = df[['Date', 'Open', 'High', 'Low', 'Close', 'Adj Close', 'Volume']]
                df.columns = ['datetime', 'open', 'high', 'low', 'close', 'adjclose', 'volume']
                df['datetime'] = df['datetime'].map(lambda x: datetime.datetime.strptime(x, '%Y-%m-%d'))
                df['index_id'] = index_i
                df['total_turnover'] = float('Nan')
                df['ret'] = df['adjclose'] / df['adjclose'].shift(1) - 1
                df = df.drop(['adjclose'], axis=1)
                df = df[(df['datetime'] >= start_date) & (df['datetime'] <= end_date)].sort_values('datetime')
                if not df.empty:
                    df_list.append(df)
            if df_list:
                df_all = pd.concat(df_list)
                self._data_helper._upload_raw(df_all, 'yahoo_index_price')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def download_all(self, start_date, end_date):
        failed_tasks = []
        if not self.cm_index_price(start_date, end_date):
            failed_tasks.append('cm_index_price')

        if not self.cxindex_index_price(start_date, end_date):
            failed_tasks.append('cxindex_index_price')

        if not self.yahoo_index_price(start_date, end_date):
            failed_tasks.append('yahoo_index_price')

        return failed_tasks


if __name__ == "__main__":
    data_helper = RawDataHelper()
    web_downloader = WebRawDataDownloader(data_helper)
    # web_downloader.cm_index_price('20200408', '20200508')
    # web_downloader.yahoo_index_price('20200427', '20200710')
